/*	$LAAS: t3dconvert2matrix.c,v 1.2 2005/08/31 21:56:52 sara Exp $ */

/* 
 * Copyright (c) 2000-2004 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "t3d/t3d.h"
#include "t3d_internal.h"


/* Local functions --------------------------------------------------- */

static void	convertVectorToMatrix(T3D *t3d);
static void	convertEulerToMatrix(T3D *t3d);


/*
 * --- t3dConvertToMatrix -----------------------------------------------
 *
 * Convert to matrix
 */

void
t3dConvertToMatrix(T3D *t3d)
{
   switch(t3d->type) {
      case T3D_VECTOR: convertVectorToMatrix(t3d);	break;
      case T3D_EULER:  convertEulerToMatrix(t3d);	break;
      case T3D_MATRIX:					break;

      default:
	 t3d->any.err = T3D_ENOTIMPLEMENTED;
	 break;
   }
}


/*
 * --- t3dMatrixOfEuler -------------------------------------------------
 *
 * (t3dMatrixOfEuler was previously t3deul.c:t3deul_2matrix)
 *
 * Compute homogeneous matrix from Euler parameters.
 *
 */

void
t3dMatrixOfEuler(double *euler, double *mat)
{
   double ca, sa, cb, sb, cg, sg;

   ca = cos(euler[0]); sa = sin(euler[0]);
   cb = cos(euler[1]); sb = sin(euler[1]);
   cg = cos(euler[2]); sg = sin(euler[2]);

   mat[0]  = ca*cb;
   mat[1]  = ca*sb*sg - sa*cg;
   mat[2]  = ca*sb*cg + sa*sg;
   mat[3]  = euler[3];

   mat[4]  = sa*cb;
   mat[5]  = sa*sb*sg + ca*cg;
   mat[6]  = sa*sb*cg - ca*sg;
   mat[7]  = euler[4];

   mat[8]  = -sb;
   mat[9]  = cb*sg;
   mat[10] = cb*cg;
   mat[11] = euler[5];

   mat[12] = 0.0;
   mat[13] = 0.0;
   mat[14] = 0.0;
   mat[15] = 1.0;
}


/*
 * --- convertEulerToMatrix ---------------------------------------------
 *
 * Convert Euler t3d to matrix
 */

static void
convertEulerToMatrix(T3D *t3d)
{
   double mat[16], cov[21];

   memcpy(cov, t3d->euler.v, 21*sizeof(double));
   t3dMatrixOfEuler(t3d->euler.euler, mat);

   memcpy(t3d->matrix.matrix, mat, sizeof(mat));
   memcpy(t3d->matrix.v, cov, sizeof(cov));
   t3d->type = T3D_MATRIX;
}


/*
 * --- convertVectorToMatrix --------------------------------------------
 *
 * Convert vector T3D to matrix
 */

static void
convertVectorToMatrix(T3D *t3d)
{
   extern const double t3dmid3[9];

   double rx = t3d->vector.vector[0];
   double ry = t3d->vector.vector[1];
   double rz = t3d->vector.vector[2];

   double m[16], cov[21];
   double dRx[9], dRy[9], dRz[9];
   double JeaR[27], Jeav[9], JRv[36];
   double s, d, d2;

   /*
    * Homogeneous matrix
    */

   d = sqrt(rx*rx + ry*ry + rz*rz);
   /* if teta = 0, the matrix is set to identity */
   if ( d > 1e-10 ) {
      double f = sin(d) / d;
      double co = cos(d);
      double g = (1 - co) / d / d;

      t3d_dm_fill(
	 m, 4, 4,
	 co+rx*rx*g,   rx*ry*g-rz*f, rx*rz*g+ry*f, t3d->vector.vector[3],
	 rx*ry*g+rz*f, co+ry*ry*g,   ry*rz*g-rx*f, t3d->vector.vector[4],
	 rx*rz*g-ry*f, ry*rz*g+rx*f, co+rz*rz*g,   t3d->vector.vector[5],
	 0.0,          0.0,          0.0,          1.0); 

   } else {
      t3d_dm_fill(
	 m, 4, 4,
	 1.0, 0.0, 0.0, t3d->vector.vector[3],
	 0.0, 1.0, 0.0, t3d->vector.vector[4],
	 0.0, 0.0, 1.0, t3d->vector.vector[5],
	 0.0, 0.0, 0.0, 1.0);
    }

  
   /*
    * Variance matrix
    */

   /* Drives angles d'Euler / termes de la matrice */
   d2 = m[0]*m[0] + m[4]*m[4];
   d = sqrt(d2);
   s = m[10]*m[10]+m[9]*m[9];

   t3d_dm_fill(
      JeaR, 3, 9,
      -m[4]/d2,    0.0, 0.0, m[0]/d2,     0.0, 0.0, 0.0, 0.0,     0.0,
      m[8]*m[0]/d, 0.0, 0.0, m[8]*m[4]/d, 0.0, 0.0, -d,  0.0,     0.0,
      0.0,             0.0, 0.0, 0.0,     0.0, 0.0, 0.0, m[10]/s, -m[9]/s);
   
   /* Drives de la matrice de rotation par raport au vecteur rotation */
   t3dDerivativeOfVector(t3d->vector.vector, dRx, dRy, dRz);
   t3d_dm_blocfill(JRv, 9, 3, 9, 1, dRx, dRy, dRz);
   
   /* The resulting jacobian matrix */
   t3d_dm_mult(JeaR, JRv, Jeav, 3, 9, 9, 3);
   t3d_dm_blocfill(JRv, 6, 6, 3, 3, Jeav, NULL, NULL, t3dmid3);
   t3d_dmfsf_jmult_6x6(JRv, t3d->vector.v, cov);
   
   /* copy result */
   memcpy(t3d->matrix.matrix, m, sizeof(m));
   memcpy(t3d->matrix.v, cov, sizeof(cov));
   t3d->type = T3D_MATRIX;
}
